#! /usr/bin/python3

"""
Run nosetest on two different branches of the same repo and compare the
results.
Usage:
nosecomp [-f|--force] [previous [current]] 

If "previous" is not given, it defaults to "master". If "current" is not given,
it defaults to the current checked out commit.

"--force" means that tests will be rerun even if there already were cached
results for one or both of the two commits.

Any uncommitted change in the working tree is neglected.
"""

__version__ = '0.1'

import shutil
import tempfile
from os import getenv
from os.path import join, exists, realpath, basename
from os import makedirs, chdir, remove
from subprocess import call, Popen, DEVNULL
from time import sleep
import argparse

from xdg.BaseDirectory import xdg_cache_home
from git import Repo

DESC = 'Compare test runs across git commits.'
parser = argparse.ArgumentParser(description=DESC)
parser.add_argument("current", nargs='?', default=None)
parser.add_argument("other", nargs='?', default='master')
# Not implemented:
#parser.add_argument("-f", "--force", dest='forced_hash', nargs="?",
#                    default=set(),
#                    help="Rerun test for given hash even if old results are "
#                         "present for it.")
parser.add_argument("-F", "--force-all", dest='force', action='store_true',
                    help="Rerun tests for all hashes even if some results are "
                         "already present.")

args = parser.parse_args()

test_command = getenv('TESTCMD', 'nosetests3')

cache_dir = join(xdg_cache_home, 'nostrils')
results_dir = join(cache_dir, 'results', test_command)
if not exists(results_dir):
    makedirs(results_dir)

class Commit(str):
    def path_out(self):
        return join(results_dir, self)

def build_exts(interpreter, dirs):
    print("Building extensions with %s" % interpreter)
    pops = []
    for a_dir in dirs:
        chdir(a_dir)
        pops.append(Popen([interpreter, 'setup.py', 'build_ext', '--inplace'],
                          stderr=DEVNULL, stdout=DEVNULL))
        chdir('..')
    while not all([p.poll() != None for p in pops]):
        sleep(.1)
    print("Finished building extensions")

def run_tests(dirs, hashes, origpath):
    pops = []
    print("Running tests on", hashes)
    fouts = [None, None]
    tempdirs = []
    for idx in range(len(dirs)):
        chdir(dirs[idx])
        fout = tempfile.NamedTemporaryFile(delete=False)
        fouts[idx] = fout
        tempdir = tempfile.mkdtemp()
        tempdirs.append(tempdir)
        # Run with ad-hoc temporary directory, to avoid the two different
        # "nosetests" processes to mess with files in '/tmp' (the environmental
        # variable is respect by the "tempfile" module)
        pops.append(Popen([test_command], stderr=fout,
                          env={"TMPDIR" : tempdir}))
        chdir('..')
    while not all([p.poll() != None for p in pops]):
        sleep(.1)
    print("Finished running tests")
    for idx, fout in enumerate(fouts):
        if fout is None:
            continue
        fout.close()
        with open(hashes[idx].path_out(), 'w') as fdef:
            with open(fout.name) as ftemp:
                fdef.write(ftemp.read().replace(dirs[idx], origpath))
        remove(fout.name)
    for tempdir in tempdirs:
        shutil.rmtree(tempdir)

def print_diff(outfiles):
    call(['git', 'diff', '--no-index', '--color-words'] + outfiles)

def clean_dirs(dirs):
    for a_dir in dirs:
        shutil.rmtree(a_dir)

def main():
    
    repo = Repo('.')
    refs = args.other, args.current
    print("Nostriling \"%s\" on %s" % (test_command, refs))
    hashes = [Commit(repo.commit(ref).hexsha) for ref in refs]
    print(hashes)
    assert(hashes[0] != hashes[1]), "The two given references coincide."
    todo = []
    if args.force:
        todo = list(hashes)
    else:
        for a_hash in hashes:
            if exists(a_hash.path_out()):
                print(a_hash, "does not need to be retested")
            else:
                todo.append(a_hash)

    if todo:
        refs_map = {repo.commit(ref).hexsha : ref for ref in refs}
        repo_path = realpath('.')
        repo_name = basename(repo_path)
        chdir(cache_dir)
        dirnames = []
        checked_out = repo.commit().hexsha
        for a_hash in todo:
            print("Preparing", a_hash)
            dirname = realpath("%s_%s" % (repo_name, a_hash))
            dirnames.append(dirname)
            if exists(dirname):
                # A previous run was aborted or failed:
                shutil.rmtree(dirname)
            shutil.copytree(repo_path, dirname)
            print("Move to", dirname)
            chdir(dirname)
            print("Moved to", dirname)
            if a_hash != checked_out:
                print("Checkout", a_hash)
                call(['git', 'checkout', refs_map[a_hash]])
                print("Checked out", refs_map[a_hash])
#            call(['git', 'reset', '--hard'])
            chdir("..")

    
        build_exts('python', dirnames)
        build_exts('python3', dirnames)
        try:
            run_tests(dirnames, todo, repo_path)
        finally:
            clean_dirs(dirnames)

    print_diff([h.path_out() for h in hashes])

if __name__ == '__main__':
    main()
